"""
Test Turtle generation logic
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2017, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import os
import unittest
import traceback

from django.test.client             import Client

from rdflib                         import Graph, URIRef, Literal

from utils.py3porting               import urljoin
from utils.SuppressLoggingContext   import SuppressLogging
from miscutils.MockHttpResources    import MockHttpFileResources, MockHttpDictResources

import annalist
from annalist                       import layout
from annalist.identifiers           import makeNamespace, RDF, RDFS, ANNAL
from annalist.util                  import make_resource_url, extract_entity_id

from annalist.models.site           import Site
from annalist.models.sitedata       import SiteData
from annalist.models.collection     import Collection
from annalist.models.recordtype     import RecordType
from annalist.models.recordview     import RecordView
from annalist.models.recordfield    import RecordField
from annalist.models.recordtypedata import RecordTypeData
from annalist.models.recordvocab    import RecordVocab
from annalist.models.annalistuser   import AnnalistUser
from annalist.models.recordenum     import RecordEnumFactory
from annalist.models.entitydata     import EntityData
from annalist.models.entitytypeinfo import EntityTypeInfo

from .AnnalistTestCase import AnnalistTestCase
from .tests import (
    TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
    )
from .init_tests import (
    init_annalist_test_site, 
    init_annalist_test_coll,
    resetSitedata
    )
from .entity_testutils import (
    site_dir, collection_dir,
    site_view_url, collection_view_url,
    collection_resource_url,
    collection_entity_view_url,
    collection_create_values,
    create_test_user
    )
from .entity_testcolldata import (
    collectiondata_url
    )
from .entity_testtypedata import (
    recordtype_url
    )
from .entity_testentitydata import (
    entity_url, entity_resource_url, entity_uriref,
    entitydata_list_type_url, entitydata_list_all_url
    )
from .entity_testsitedata import (
    get_site_types, get_site_types_sorted, get_site_types_linked,
    )

#   -----------------------------------------------------------------------------
#
#   RDF graph utilities
#
#   -----------------------------------------------------------------------------

def property_value(graph, item, prop):
    """
    Returns property value associated with a node in a graph, or None
    """
    return graph.value(item, URIRef(prop))

def scan_list(graph, head):
    """
    Returns an iterator over a list whose head node is supplied.
    """
    while head != URIRef(RDF.URI.nil):
        yield property_value(graph, head, RDF.URI.first)
        head = property_value(graph, head, RDF.URI.rest)
    return

#   -----------------------------------------------------------------------------
#
#   Test data
#
#   -----------------------------------------------------------------------------

_test_test_vocab_create_values = (
    { "annal:type":     "annal:Vocabulary"
    , "rdfs:label":     "Vocabulary namespace for test terms"
    , "rdfs:comment":   "Vocabulary namespace for URIs that are used internally by Annalist to identify application types and properties."
    , "annal:uri":      "http://example.org/test/#"
    })

test_blob_vocab_create_values = (
    { "annal:type":     "annal:Vocabulary"
    , "rdfs:label":     "Vocabulary namespace for test terms"
    , "rdfs:comment":   "Vocabulary namespace for URIs that are used internally by Annalist to identify application types and properties."
    , "annal:uri":      "http://example.org/blob/yyy#"
    })

test_image_ref_type_create_values = (
    { 'annal:type':                 "annal:Type"
    , 'rdfs:label':                 "test_reference_type label"
    , 'rdfs:comment':               "test_reference_type comment"
    , 'annal:uri':                  "blob:type/test_reference_type"
    , 'annal:type_view':            "test_reference_view"
    , 'annal:type_list':            "test_reference_list"
    })

test_image_ref_view_create_values = (
    { 'annal:type':                 "annal:View"
    , 'rdfs:label':                 "test_image_view label"
    , 'rdfs:comment':               "test_image_view comment"
    , 'annal:view_entity_type':     ""
    , 'annal:view_fields':
      [ { 'annal:field_id':             "Entity_id"
        , 'annal:field_placement':      "small:0,12;medium:0,6"
        }
      , { 'annal:field_id':             "Entity_label"
        , 'annal:field_placement':      "small:0,12"
        }
      , { 'annal:field_id':             "Entity_comment"
        , 'annal:field_placement':      "small:0,12"
        }
      , { 'annal:field_id':             "Test_image_ref"
        , 'annal:field_placement':      "small:0,12"
        }
      ]
    })

test_image_ref_field_create_values = (
    { 'annal:type':                     "annal:Field"
    , 'annal:field_name':               "ref_image"
    , 'rdfs:label':                     "test_image_ref_field label"
    , 'rdfs:comment':                   "test_image_ref_field comment"
    , 'annal:property_uri':             "blob:reference"
    , 'annal:field_render_type':        "RefImage"
    , 'annal:field_value_mode':         "Value_direct"
    , 'annal:field_value_type':        "annal:Identifier"
    , 'annal:placeholder':              "(Image reference)"
    , 'annal:default_value':            ""
    })

def test_ref_entity_create_values(image_uri):
    return (
        { 'rdfs:label':                 "test_ref_image label"
        , 'rdfs:comment':               "test_ref_image comment"
        , 'blob:reference':             image_uri
        })

#   -----------------------------------------------------------------------------
#
#   Turtle generation tests
#
#   -----------------------------------------------------------------------------

class TurtleOutputTest(AnnalistTestCase):
    """
    Tests site/collection data interpreted as Turtle
    """

    def setUp(self):
        self.testsite    = init_annalist_test_site()
        self.testcoll    = init_annalist_test_coll()
        self.sitebasedir = TestBaseDir
        self.collbasedir = os.path.join(
            self.sitebasedir, layout.SITEDATA_DIR, layout.COLL_BASE_DIR
            )
        # Login and permissions
        create_test_user(self.testcoll, "testuser", "testpassword")
        self.client = Client(HTTP_HOST=TestHost)
        loggedin = self.client.login(username="testuser", password="testpassword")
        self.assertTrue(loggedin)
        return

    def tearDown(self):
        # resetSitedata(scope="collections")
        return

    @classmethod
    def setUpClass(cls):
        super(TurtleOutputTest, cls).setUpClass()
        return

    @classmethod
    def tearDownClass(cls):
        super(TurtleOutputTest, cls).tearDownClass()
        resetSitedata()
        return

    #   -----------------------------------------------------------------------------
    #   Helpers
    #   -----------------------------------------------------------------------------

    def coll_basedir(self, coll_id):
        return os.path.normpath(os.path.join(self.collbasedir, "../../%s/d/"%coll_id))

    def entity_basedir(self, coll_id, type_id, entity_id):
        return os.path.join(
            self.sitebasedir, 
            layout.SITE_ENTITY_PATH%{'coll_id': coll_id, 'type_id': type_id, 'id': entity_id}
            )

    def dir_base_url(self, b):
        return "file://"+b+"/"

    def coll_baseurl(self, coll_id):
        return self.dir_base_url(self.coll_basedir(coll_id))

    def entity_baseurl(self, coll_id, type_id, entity_id):
        return self.dir_base_url(self.entity_basedir(coll_id, type_id, entity_id))

    def coll_url(self, coll_id):
        return urljoin(self.coll_baseurl(coll_id), layout.META_COLL_REF)

    def entity_url(self, coll_id, type_id, entity_id):
        return "file://" + self.entity_basedir(coll_id, type_id, entity_id)

    def resolve_coll_url(self, coll, ref):
        coll_base    = urljoin(self.testcoll.get_url(), layout.COLL_BASE_REF)
        resolved_url = urljoin(coll_base, ref)
        return resolved_url

    def scan_rdf_list(self, graph, head):
        """
        Iterate over nodes in an RDF list
        """
        next = head
        while (next is not None) and (next != URIRef(RDF.URI.nil)):
            item = graph.value(subject=next, predicate=URIRef(RDF.URI.first))
            yield item
            next = graph.value(subject=next, predicate=URIRef(RDF.URI.rest))
        return

    def assertTripleIn(self, t, g):
        """
        Assert that triple exists in graph, where None values are wildcard
        """
        ts = list(g.triples(t))
        self.assertNotEqual(len(ts), 0, "Triple %r not in graph"%(t,))
        return

    def get_context_mock_dict(self, base_path, context_path="../../"):
        """
        Uses Django test client results to create a dictionary of mock results for 
        accessing JSONLD context resources.  Works with MockHttpDictResources.
        """
        mock_refs = (
            [ context_path+"coll_context.jsonld"
            ])
        mock_dict = {}
        for mock_ref in mock_refs:
            mu = urljoin(base_path, mock_ref)
            log.debug(
                "get_context_mock_dict: base_path %s, mock_ref %s, mu %s"%
                (base_path, mock_ref, mu)
                )
            mr = self.client.get(mu)
            if mr.status_code != 200:
                log.error(
                    "get_context_mock_dict: uri %s, status_code %d, reason_phrase %s"%
                    (mu, mr.status_code, mr.reason_phrase)
                    )
                # log.error("".join(traceback.format_stack()))
            self.assertEqual(mr.status_code,   200)
            mock_dict[mock_ref] = mr.content
        # print "***** get_context_mock_dict: mu: %s, mock_dict: %r"%(mu, mock_dict.keys())
        return mock_dict

    #   -----------------------------------------------------------------------------
    #   Turtle output tests
    #   -----------------------------------------------------------------------------

    # Site data view (check data links)
    def test_get_site_data_view(self):
        site_url = site_view_url()
        site_url = collection_view_url(layout.SITEDATA_ID)
        u = collectiondata_url(coll_id=layout.SITEDATA_ID)
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertEqual(r.context['coll_id'],          layout.SITEDATA_ID)
        self.assertEqual(r.context['type_id'],          layout.COLL_TYPEID)
        self.assertEqual(r.context['entity_id'],        layout.SITEDATA_ID)
        self.assertEqual(r.context['orig_id'],          layout.SITEDATA_ID)
        self.assertEqual(r.context['action'],           "view")
        self.assertEqual(r.context['continuation_url'], "")
        self.assertEqual(
            r.context['entity_data_ref'],      
            site_url+layout.COLL_META_REF
            )
        self.assertEqual(
            r.context['entity_turtle_ref'], 
            site_url+layout.COLL_TURTLE_REF
            )
        return

    # Site data content negotiation
    def test_get_site_data_turtle(self):
        """
        Request site data as Turtle
        """
        site_url = site_view_url()
        site_url = collection_view_url(layout.SITEDATA_ID)
        u = collectiondata_url(coll_id=layout.SITEDATA_ID)
        r = self.client.get(u, HTTP_ACCEPT="text/turtle")
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        v = r['Location']
        self.assertEqual(v, TestHostUri+site_url+layout.COLL_TURTLE_REF)
        w = site_url + layout.COLL_BASE_REF
        mock_resource_dict = self.get_context_mock_dict(w, layout.META_COLL_BASE_REF)
        with MockHttpDictResources(v, mock_resource_dict):
            r = self.client.get(v)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        return

    def test_head_site_data_turtle(self):
        """
        Request HEAD for site data as Turtle
        """
        site_url = site_view_url()
        site_url = collection_view_url(layout.SITEDATA_ID)
        u = collectiondata_url(coll_id=layout.SITEDATA_ID)
        r = self.client.head(u, HTTP_ACCEPT="text/turtle")
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        v = r['Location']
        self.assertEqual(v, TestHostUri+site_url+layout.COLL_TURTLE_REF)
        return

    def test_http_turtle_site(self):
        """
        Read site data as Turtle, and check resulting RDF triples
        """
        # Read collection data as Turtle
        site_url = site_view_url()
        site_url = collection_view_url(layout.SITEDATA_ID)
        v = site_url + layout.COLL_BASE_REF
        u = TestHostUri + site_url + layout.COLL_TURTLE_REF
        mock_resource_dict = self.get_context_mock_dict(v, layout.META_COLL_BASE_REF)
        with MockHttpDictResources(u, mock_resource_dict):
            r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Parse data as Turtle
        g = Graph()
        result = g.parse(data=r.content, publicID=u, format="turtle")
        # print "*****"+repr(result)
        # print "***** site:"
        # print(g.serialize(format='turtle', indent=4))

        # Check the resulting graph contents
        subj             = TestHostUri + site_url
        site_data        = self.testsite.site_data()
        ann_id           = Literal(layout.SITEDATA_ID)
        ann_type         = URIRef(ANNAL.URI.SiteData)
        ann_type_id      = Literal(SiteData._entitytypeid)
        software_version = Literal(annalist.__version_data__)
        label            = Literal(site_data[RDFS.CURIE.label])
        comment          = Literal(site_data[RDFS.CURIE.comment])
        for (s, p, o) in (
            [ (subj, RDFS.URI.label,             label              )
            , (subj, RDFS.URI.comment,           comment            )
            , (subj, ANNAL.URI.id,               ann_id             )
            , (subj, ANNAL.URI.type,             ann_type           )
            , (subj, ANNAL.URI.type_id,          ann_type_id        )
            , (subj, ANNAL.URI.software_version, software_version   )
            ]):
            self.assertIn( (URIRef(s), URIRef(p), o), g)
        return

    def test_http_turtle_collection(self):
        """
        Read new collection data as Turtle, and check resulting RDF triples
        """
        # Generate collection JSON-LD context data
        self.testcoll.generate_coll_jsonld_context()
        # Read collection data as Turtle
        collection_url = collection_view_url(coll_id="testcoll")
        v = collection_url + layout.COLL_BASE_REF
        u = TestHostUri + collection_url + layout.COLL_TURTLE_REF
        mock_resource_dict = self.get_context_mock_dict(v, layout.META_COLL_BASE_REF)
        with MockHttpDictResources(u, mock_resource_dict):
            r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Parse data as Turtle
        g = Graph()
        result = g.parse(data=r.content, publicID=u, format="turtle")
        # print "*****"+repr(result)
        # print "***** coll:"
        # print(g.serialize(format='turtle', indent=4))
        # Check the resulting graph contents
        subj      = TestHostUri + collection_url
        coll_data = self.testcoll._load_values()
        for (s, p, o) in (
            [ (subj, RDFS.URI.label,             Literal(coll_data[RDFS.CURIE.label])       )
            , (subj, RDFS.URI.comment,           Literal(coll_data[RDFS.CURIE.comment])     )
            , (subj, ANNAL.URI.id,               Literal(coll_data[ANNAL.CURIE.id])         )
            , (subj, ANNAL.URI.type_id,          Literal(coll_data[ANNAL.CURIE.type_id])    )
            , (subj, ANNAL.URI.type,             URIRef(ANNAL.URI.Collection)               )
            , (subj, ANNAL.URI.software_version, Literal(annalist.__version_data__)         )
            ]):
            self.assertIn( (URIRef(s), URIRef(p), o), g)
        return




    # @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@







    # Collection data view
    def test_get_collection_data_view(self):
        collection_url = collection_view_url(coll_id="testcoll")
        u = collectiondata_url(coll_id="testcoll")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertEqual(r.context['coll_id'],          layout.SITEDATA_ID)
        self.assertEqual(r.context['type_id'],          layout.COLL_TYPEID)
        self.assertEqual(r.context['entity_id'],        "testcoll")
        self.assertEqual(r.context['orig_id'],          "testcoll")
        self.assertEqual(r.context['action'],           "view")
        self.assertEqual(r.context['continuation_url'], "")
        self.assertEqual(
            r.context['entity_data_ref'],      
            collection_url+layout.COLL_META_REF
            )
        self.assertEqual(
            r.context['entity_turtle_ref'], 
            collection_url+layout.COLL_TURTLE_REF
            )
        return

    # Collection data content negotiation
    def test_get_collection_data_turtle(self):
        """
        Request collection data as Turtle
        """
        collection_url = collection_view_url(coll_id="testcoll")
        u = collectiondata_url(coll_id="testcoll")
        r = self.client.get(u, HTTP_ACCEPT="text/turtle")
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        v = r['Location']
        self.assertEqual(v, TestHostUri+collection_url+layout.COLL_TURTLE_REF)
        w = collection_url + layout.COLL_BASE_REF
        mock_resource_dict = self.get_context_mock_dict(w, layout.META_COLL_BASE_REF)
        with MockHttpDictResources(v, mock_resource_dict):
            r = self.client.get(v)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        return

    def test_http_turtle_collection(self):
        """
        Read new collection data as Turtle, and check resulting RDF triples
        """
        # Generate collection JSON-LD context data
        self.testcoll.generate_coll_jsonld_context()
        # Read collection data as Turtle
        collection_url = collection_view_url(coll_id="testcoll")
        v = collection_url + layout.COLL_BASE_REF
        u = TestHostUri + collection_url + layout.COLL_TURTLE_REF
        mock_resource_dict = self.get_context_mock_dict(v, layout.META_COLL_BASE_REF)
        with MockHttpDictResources(u, mock_resource_dict):
            r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Parse data as Turtle
        g = Graph()
        result = g.parse(data=r.content, publicID=u, format="turtle")
        # print "*****"+repr(result)
        # print "***** coll:"
        # print(g.serialize(format='turtle', indent=4))
        # Check the resulting graph contents
        subj      = TestHostUri + collection_url
        coll_data = self.testcoll._load_values()
        for (s, p, o) in (
            [ (subj, RDFS.URI.label,             Literal(coll_data[RDFS.CURIE.label])       )
            , (subj, RDFS.URI.comment,           Literal(coll_data[RDFS.CURIE.comment])     )
            , (subj, ANNAL.URI.id,               Literal(coll_data[ANNAL.CURIE.id])         )
            , (subj, ANNAL.URI.type_id,          Literal(coll_data[ANNAL.CURIE.type_id])    )
            , (subj, ANNAL.URI.type,             URIRef(ANNAL.URI.Collection)               )
            , (subj, ANNAL.URI.software_version, Literal(annalist.__version_data__)         )
            ]):
            self.assertIn( (URIRef(s), URIRef(p), o), g)
        return

    def test_turtle_enum_list_type_grid(self):
        """
        Read enumeration value as Turtle, and check resulting RDF triples
        """
        # Generate collection JSON-LD context data
        self.testcoll.generate_coll_jsonld_context()
        Enum_list_type = RecordEnumFactory(layout.ENUM_LIST_TYPE_ID, layout.ENUM_LIST_TYPE_ID)
        list_type_grid = Enum_list_type.load(
            self.testcoll, "Grid", altscope="all"
            )
        # Read enumerated value list data as Turtle
        type_id = list_type_grid.get_type_id()
        enum_id = list_type_grid.get_id()
        v = entity_url(coll_id="testcoll", type_id=type_id, entity_id=enum_id)
        u = TestHostUri + entity_resource_url(
            coll_id="testcoll", type_id=type_id, entity_id=enum_id,
            resource_ref=layout.VOCAB_META_TURTLE
            )
        # print("@@@@ v: (list_type_grid)")
        # print(repr(v))
        # print("@@@@ u: (list_type_grid)")
        # print(repr(u))
        with MockHttpDictResources(u, self.get_context_mock_dict(v)):
            r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Check the resulting graph contents
        g = Graph()
        result = g.parse(data=r.content, publicID=u, format="turtle")
        # print("@@@@ g: (list_type_grid)")
        # print(g.serialize(format='turtle', indent=4))
        subj                = TestHostUri + v.rstrip("/")
        list_type_grid_data = list_type_grid.get_values()
        list_type_url       = ANNAL.to_uri(list_type_grid_data[ANNAL.CURIE.uri])
        for (s, p, o) in (
            [ (subj, RDF.URI.type,          URIRef(ANNAL.URI.Enum)                            )
            , (subj, RDF.URI.type,          URIRef(ANNAL.URI.Enum_list_type)                  )
            , (subj, RDFS.URI.label,        Literal(list_type_grid_data[RDFS.CURIE.label])    )
            , (subj, RDFS.URI.comment,      Literal(list_type_grid_data[RDFS.CURIE.comment])  )
            , (subj, ANNAL.URI.id,          Literal(list_type_grid_data[ANNAL.CURIE.id])      )
            , (subj, ANNAL.URI.type_id,     Literal(list_type_grid_data[ANNAL.CURIE.type_id]) )
            , (subj, ANNAL.URI.uri,         URIRef(list_type_url)                             )
            ]):
            self.assertIn( (URIRef(s), URIRef(p), o), g )
        return

    def test_http_turtle_entity1(self):
        """
        Read default entity data as Turtle using HTTP, and check resulting RDF triples
        """
        # Generate collection JSON-LD context data
        self.testcoll.generate_coll_jsonld_context()
        # Read entity data as Turtle, with HTTP mocking for context references
        v = entity_url(coll_id="testcoll", type_id="testtype", entity_id="entity1")
        u = TestHostUri + entity_resource_url(
            coll_id="testcoll", type_id="testtype", entity_id="entity1",
            resource_ref=layout.ENTITY_DATA_TURTLE
            )
        with MockHttpDictResources(u, self.get_context_mock_dict(v)):
            r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Parse data as Turtle
        g = Graph()
        result = g.parse(data=r.content, publicID=u, format="turtle")
        # Check the resulting graph contents
        subj          = TestHostUri + v.rstrip("/")
        testtype_data = RecordTypeData.load(self.testcoll, "testtype")
        entity1       = EntityData.load(testtype_data, "entity1")
        entity_data   = entity1.get_values()
        for (s, p, o) in (
            [ (subj, RDFS.URI.label,     Literal(entity_data[RDFS.CURIE.label])    )
            , (subj, RDFS.URI.comment,   Literal(entity_data[RDFS.CURIE.comment])  )
            , (subj, ANNAL.URI.id,       Literal(entity_data[ANNAL.CURIE.id])      )
            , (subj, ANNAL.URI.type_id,  Literal(entity_data[ANNAL.CURIE.type_id]) )
            , (subj, ANNAL.URI.type,     URIRef(ANNAL.URI.EntityData)              )
            ]):
            self.assertIn( (URIRef(s), URIRef(p), o), g)
        return

    def test_http_turtle_type_vocab(self):
        """
        Read type data as Turtle, and check resulting RDF triples
        """
        # Generate collection JSON-LD context data
        self.testcoll.generate_coll_jsonld_context()
        # Read type data as Turtle, with HTTP mocking for context references
        type_vocab = self.testcoll.get_type("_vocab")
        v = recordtype_url(coll_id="testcoll", type_id=type_vocab.get_id())
        p = v + layout.TYPE_META_TURTLE
        u = TestHostUri + p
        with MockHttpDictResources(u, self.get_context_mock_dict(p)):
            r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Parse graph data as Turtle
        g = Graph()
        result = g.parse(data=r.content, publicID=u, format="turtle")
        # Check the resulting graph contents
        subj            = TestHostUri + v.rstrip("/")
        type_vocab_data = type_vocab.get_values()
        vocab_list_url  = self.resolve_coll_url(self.testcoll, type_vocab_data[ANNAL.CURIE.type_list])
        vocab_view_url  = self.resolve_coll_url(self.testcoll, type_vocab_data[ANNAL.CURIE.type_view])
        for (s, p, o) in (
            [ (subj, RDF.URI.type,        URIRef(ANNAL.URI.Type)                        )
            , (subj, RDFS.URI.label,      Literal(type_vocab_data[RDFS.CURIE.label])    )
            , (subj, RDFS.URI.comment,    Literal(type_vocab_data[RDFS.CURIE.comment])  )
            , (subj, ANNAL.URI.id,        Literal(type_vocab_data[ANNAL.CURIE.id])      )
            , (subj, ANNAL.URI.type_id,   Literal(type_vocab_data[ANNAL.CURIE.type_id]) )
            , (subj, ANNAL.URI.type_list, URIRef(vocab_list_url)                        )
            , (subj, ANNAL.URI.type_view, URIRef(vocab_view_url)                        )
            , (subj, ANNAL.URI.uri,       URIRef(ANNAL.URI.Vocabulary)                  )
            ]):
            self.assertIn( (URIRef(s), URIRef(p), o), g )
        return

    def test_http_turtle_type_new(self):
        """
        Create new type in collection, then read type data as Turtle, 
        and check resulting RDF triples
        """
        # Create new type in collection
        test_new_type_create_values = (
            { 'annal:type':                 "annal:Type"
            , 'rdfs:label':                 "test_new_type label"
            , 'rdfs:comment':               "test_new_type comment"
            , 'annal:uri':                  "blob:type/test_new_type"
            , 'annal:type_view':            "Default_view"
            , 'annal:type_list':            "Default_list"
            })
        new_type = RecordType.create(
            self.testcoll, "newtype", test_new_type_create_values
            )
        blob_vocab = RecordVocab.create(
            self.testcoll, "blob", test_blob_vocab_create_values
            )
        # Generate collection JSON-LD context data
        self.testcoll.generate_coll_jsonld_context()
        # Read type data as Turtle, with HTTP mocking for context references
        type_new = self.testcoll.get_type("newtype")
        v = recordtype_url(coll_id="testcoll", type_id=type_new.get_id())
        p = v + layout.TYPE_META_TURTLE
        u = TestHostUri + p
        with MockHttpDictResources(u, self.get_context_mock_dict(p)):
            r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Parse graph data as Turtle
        g = Graph()
        result = g.parse(data=r.content, publicID=u, format="turtle")
        # print "*****"+repr(result)
        # print("***** g: (type_new_data)")
        # print(g.serialize(format='turtle', indent=4))
        # Check the resulting graph contents
        subj          = TestHostUri + v.rstrip("/")
        type_new_data = type_new.get_values()
        new_list_url  = self.resolve_coll_url(self.testcoll, type_new_data[ANNAL.CURIE.type_list])
        new_view_url  = self.resolve_coll_url(self.testcoll, type_new_data[ANNAL.CURIE.type_view])
        # type_new_uri  = type_new_data[ANNAL.CURIE.uri]  # "blob:type/test_new_type"
        blobns         = makeNamespace("blob", "http://example.org/blob/yyy#", ["reference"])
        type_new_uri   = blobns.to_uri(type_new_data[ANNAL.CURIE.uri]) 
        for (s, p, o) in (
            [ (subj, RDF.URI.type,        URIRef(ANNAL.URI.Type)                        )
            , (subj, RDFS.URI.label,      Literal(type_new_data[RDFS.CURIE.label])      )
            , (subj, RDFS.URI.comment,    Literal(type_new_data[RDFS.CURIE.comment])    )
            , (subj, ANNAL.URI.id,        Literal(type_new_data[ANNAL.CURIE.id])        )
            , (subj, ANNAL.URI.type_id,   Literal(type_new_data[ANNAL.CURIE.type_id])   )
            , (subj, ANNAL.URI.type_list, URIRef(new_list_url)                          )
            , (subj, ANNAL.URI.type_view, URIRef(new_view_url)                          )
            , (subj, ANNAL.URI.uri,       URIRef(type_new_uri)                          )
            ]):
            self.assertIn( (URIRef(s), URIRef(p), o), g )
        return

    def test_http_turtle_vocab_annal(self):
        """
        Read annal: vocabulary data as Turtle using HTTP, and check resulting RDF triples
        """
        # Generate collection JSON-LD context data
        self.testcoll.generate_coll_jsonld_context()
        # Read vocabulary data as Turtle
        v = entity_url(coll_id="testcoll", type_id="_vocab", entity_id="annal")
        u = TestHostUri + entity_resource_url(
            coll_id="testcoll", type_id="_vocab", entity_id="annal",
            resource_ref=layout.VOCAB_META_TURTLE
            )
        with MockHttpDictResources(u, self.get_context_mock_dict(v)):
            r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")

        # Parse vocabulary data as Turtle
        g = Graph()
        result = g.parse(data=r.content, publicID=u, format="turtle")

        # Check the resulting graph contents
        subj          = TestHostUri + v.rstrip("/")
        vocabtypeinfo = EntityTypeInfo(self.testcoll, "_vocab")
        annalvocab    = vocabtypeinfo.get_entity("annal")
        annal_data    = annalvocab.get_values()
        seeAlso_1     = "https://github.com/gklyne/annalist/blob/master/src/annalist_root/annalist/identifiers.py"
        for (s, p, o) in (
            [ (subj, RDF.URI.type,      URIRef(ANNAL.URI.EntityData)             )
            , (subj, RDF.URI.type,      URIRef(ANNAL.URI.Vocabulary)             )
            , (subj, RDFS.URI.label,    Literal(annal_data[RDFS.CURIE.label])    )
            , (subj, RDFS.URI.comment,  Literal(annal_data[RDFS.CURIE.comment])  )
            , (subj, RDFS.URI.seeAlso,  URIRef(seeAlso_1)                        )            
            , (subj, ANNAL.URI.id,      Literal(annal_data[ANNAL.CURIE.id])      )
            , (subj, ANNAL.URI.type_id, Literal(annal_data[ANNAL.CURIE.type_id]) )
            , (subj, ANNAL.URI.type,    URIRef(ANNAL.URI.Vocabulary)             )
            , (subj, ANNAL.URI.uri,     URIRef(ANNAL._baseUri)                   )
            ]):
            self.assertIn( (URIRef(s), URIRef(p), o), g)
        return

    def test_http_turtle_user_view(self):
        """
        Read user view data as Turtle, and check resulting RDF triples
        """
        # Generate collection JSON-LD context data
        self.testcoll.generate_coll_jsonld_context()

        # Read vocabulary data as Turtle
        u = TestHostUri + entity_resource_url(
            coll_id="testcoll", type_id="_view", entity_id="User_view",
            resource_ref=layout.VIEW_META_TURTLE
            )
        v = entity_url(coll_id="testcoll", type_id="_view", entity_id="User_view")
        with MockHttpDictResources(u, self.get_context_mock_dict(v)):
            r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")

        # Parse view data as Turtle
        g = Graph()
        result = g.parse(data=r.content, publicID=u, format="turtle")

        # Check the resulting graph contents
        subj           = TestHostUri + v.rstrip("/")
        viewtypeinfo   = EntityTypeInfo(self.testcoll, "_view")
        view_user      = viewtypeinfo.get_entity("User_view")
        view_user_data = view_user.get_values()
        view_uri       = ANNAL.to_uri(view_user_data[ANNAL.CURIE.uri])
        for (s, p, o) in (
            [ (subj, RDF.URI.type,          URIRef(ANNAL.URI.View)                       )
            , (subj, RDFS.URI.label,        Literal(view_user_data[RDFS.CURIE.label])    )
            , (subj, RDFS.URI.comment,      Literal(view_user_data[RDFS.CURIE.comment])  )
            , (subj, ANNAL.URI.id,          Literal(view_user_data[ANNAL.CURIE.id])      )
            , (subj, ANNAL.URI.type_id,     Literal(view_user_data[ANNAL.CURIE.type_id]) )
            , (subj, ANNAL.URI.uri,         URIRef(view_uri)                             )
            , (subj, ANNAL.URI.open_view,   Literal(False)                               )
            , (subj, ANNAL.URI.view_entity_type, URIRef(ANNAL.URI.User)                       )
            ]):
            self.assertIn( (URIRef(s), URIRef(p), o), g )
        # Check field list contents
        fields = view_user_data[ANNAL.CURIE.view_fields]
        head   = property_value(g, URIRef(subj), ANNAL.URI.view_fields)
        items  = scan_list(g, head)
        for f in fields:
            fi  = URIRef(
                TestHostUri + collection_entity_view_url(
                    type_id="_field", 
                    entity_id=extract_entity_id(f[ANNAL.CURIE.field_id])
                    ).rstrip("/")
                )
            fp  = Literal(f[ANNAL.CURIE.field_placement])
            fn  = next(items)
            fni = property_value(g, fn, ANNAL.URI.field_id)
            fnp = property_value(g, fn, ANNAL.URI.field_placement)
            self.assertEqual(fni, fi)
            self.assertEqual(fnp, fp)
        # self.assertRaises as context manager, see http://stackoverflow.com/a/28223420/324122
        with self.assertRaises(StopIteration):
            next(items)
        return

    def test_http_turtle_user_default(self):
        """
        Read default user data as Turtle, and check resulting RDF triples
        """
        # Generate collection JSON-LD context data
        self.testcoll.generate_coll_jsonld_context()

        # Read vocabulary data as Turtle
        u = TestHostUri + entity_resource_url(
            coll_id="testcoll", type_id="_user", entity_id="_default_user_perms",
            resource_ref=layout.USER_META_TURTLE
            )
        v = entity_url(coll_id="testcoll", type_id="_user", entity_id="_default_user_perms")
        with MockHttpDictResources(u, self.get_context_mock_dict(v)):
            r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")

        # Parse view data as Turtle
        g = Graph()
        result = g.parse(data=r.content, publicID=u, format="turtle")

        # Check the resulting graph contents
        subj              = TestHostUri + v.rstrip("/")
        usertypeinfo      = EntityTypeInfo(self.testcoll, "_user")
        user_default_data = usertypeinfo.get_entity("_default_user_perms")
        user_uri          = ANNAL.to_uri(user_default_data[ANNAL.CURIE.user_uri])
        user_perms        = user_default_data[ANNAL.CURIE.user_permission]
        for (s, p, o) in (
            [ (subj, RDF.URI.type,          URIRef(ANNAL.URI.User)                          )
            , (subj, RDFS.URI.label,        Literal(user_default_data[RDFS.CURIE.label])    )
            , (subj, RDFS.URI.comment,      Literal(user_default_data[RDFS.CURIE.comment])  )
            , (subj, ANNAL.URI.id,          Literal(user_default_data[ANNAL.CURIE.id])      )
            , (subj, ANNAL.URI.type_id,     Literal(user_default_data[ANNAL.CURIE.type_id]) )
            , (subj, ANNAL.URI.user_uri,    URIRef(user_uri)                                )
            , (subj, ANNAL.URI.user_permission, Literal(user_perms[0])                      )
            ]):
            self.assertIn( (URIRef(s), URIRef(p), o), g )
        return

    def test_http_turtle_image_ref(self):
        """
        Read image reference data as Turtle, and check resulting RDF triples
        """
        # Populate collection with record type, view and field
        imageurl = TestHostUri + "/test-image.jpg" 
        test_ref_type = RecordVocab.create(
            self.testcoll, "blob", test_blob_vocab_create_values
            )
        test_ref_type = RecordType.create(
            self.testcoll, "testreftype", test_image_ref_type_create_values
            )
        test_ref_view = RecordView.create(
            self.testcoll, "testrefview", test_image_ref_view_create_values
            )
        test_ref_field = RecordField.create(
            self.testcoll, "Test_image_ref", test_image_ref_field_create_values
            )
        # Create data records for testing image references:
        test_ref_type_info = EntityTypeInfo(
            self.testcoll, "testreftype", create_typedata=True
            )
        test_ref_type_info.create_entity("refentity", test_ref_entity_create_values(imageurl))
        # Generate collection JSON-LD context data
        self.testcoll.generate_coll_jsonld_context()
        # Read data as Turtle, with HTTP mocking for context references
        u = TestHostUri + entity_resource_url(
            coll_id="testcoll", type_id="testreftype", entity_id="refentity",
            resource_ref=layout.ENTITY_DATA_TURTLE
            )
        v = entity_url(coll_id="testcoll", type_id="testreftype", entity_id="refentity")
        with MockHttpDictResources(u, self.get_context_mock_dict(v)):
            r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Parse data as Turtle
        g = Graph()
        result = g.parse(data=r.content, publicID=u, format="turtle")
        # print "*****"+repr(result)
        # print("***** g: (ref_image_data)")
        # print(g.serialize(format='turtle', indent=4))
        # Check the resulting graph contents
        subj              = TestHostUri + v.rstrip("/")
        testdata       = RecordTypeData.load(self.testcoll, "testreftype")
        ref_image      = EntityData.load(testdata, "refentity")
        ref_image_data = ref_image.get_values()
        blobns         = makeNamespace("blob", "http://example.org/blob/yyy#", ["reference"])
        type_uri       = blobns.to_uri(ref_image_data['@type'][0]) 
        for (s, p, o) in (
            [ (subj, RDF.URI.type,          URIRef(ANNAL.URI.EntityData)                 )
            , (subj, RDF.URI.type,          URIRef(type_uri)                             )
            , (subj, RDFS.URI.label,        Literal(ref_image_data[RDFS.CURIE.label])    )
            , (subj, RDFS.URI.comment,      Literal(ref_image_data[RDFS.CURIE.comment])  )
            , (subj, ANNAL.URI.id,          Literal(ref_image_data[ANNAL.CURIE.id])      )
            , (subj, ANNAL.URI.type_id,     Literal(ref_image_data[ANNAL.CURIE.type_id]) )
            , (subj, blobns.URI.reference,  URIRef(imageurl)                             )
            ]):
            self.assertIn( (URIRef(s), URIRef(p), o), g )
        return

    #   -----------------------------------------------------------------------------
    #   Entity content negotiation tests
    #   -----------------------------------------------------------------------------

    def test_http_conneg_turtle_entity1(self):
        """
        Read default entity data as Turtle using HTTP, and check resulting RDF triples
        """
        # Generate collection JSON-LD context data
        self.testcoll.generate_coll_jsonld_context()
        # Create entity object to access entity data 
        testdata = RecordTypeData.load(self.testcoll, "testtype")
        entity1  = EntityData.load(testdata, "entity1")
        # Read entity data as Turtle
        u = entity_url(coll_id="testcoll", type_id="testtype", entity_id="entity1")
        # print "@@ test_http_conneg_turtle_entity1: uri %s"%u
        r = self.client.get(u, HTTP_ACCEPT="text/turtle")
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        v = r['Location']
        self.assertEqual(v, TestHostUri+u+layout.ENTITY_DATA_TURTLE)
        with MockHttpDictResources(v, self.get_context_mock_dict(v)):
            r = self.client.get(v)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        g = Graph()
        # print("***** v: (entity1)")
        # print(repr(v))
        # print("***** c: (entity1)")
        # print r.content
        result = g.parse(data=r.content, publicID=v, format="turtle")
        # print "*****"+repr(result)
        # print("***** g: (entity1)")
        # print(g.serialize(format='turtle', indent=4))
        # Check the resulting graph contents
        subj        = entity1.get_url().rstrip("/")
        entity_data = entity1.get_values()
        for (s, p, o) in (
            [ (subj, RDFS.URI.label,     Literal(entity_data[RDFS.CURIE.label])    )
            , (subj, RDFS.URI.comment,   Literal(entity_data[RDFS.CURIE.comment])  )
            , (subj, ANNAL.URI.id,       Literal(entity_data[ANNAL.CURIE.id])      )
            , (subj, ANNAL.URI.type_id,  Literal(entity_data[ANNAL.CURIE.type_id]) )
            , (subj, ANNAL.URI.type,     URIRef(ANNAL.URI.EntityData)              )
            ]):
            self.assertIn( (URIRef(s), URIRef(p), o), g)
        return

    def test_http_conneg_head_turtle_entity1(self):
        """
        HEAD request to entity with content negotiation for Turtle
        """
        # Generate collection JSON-LD context data
        self.testcoll.generate_coll_jsonld_context()
        # Create entity object to access entity data 
        testdata = RecordTypeData.load(self.testcoll, "testtype")
        entity1  = EntityData.load(testdata, "entity1")
        # Read entity data as Turtle
        u = entity_url(coll_id="testcoll", type_id="testtype", entity_id="entity1")
        r = self.client.head(u, HTTP_ACCEPT="text/turtle")
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        v = r['Location']
        self.assertEqual(v, TestHostUri+u+layout.ENTITY_DATA_TURTLE)
        return

    def test_http_conneg_turtle_type_vocab(self):
        """
        Read type data as Turtle, and check resulting RDF triples
        """
        # Generate collection JSON-LD context data
        self.testcoll.generate_coll_jsonld_context()
        type_vocab = self.testcoll.get_type("_vocab")
        # Read type data as Turtle, with HTTP mocking for context references
        u = recordtype_url(coll_id="testcoll", type_id=type_vocab.get_id())
        r = self.client.get(u, HTTP_ACCEPT="text/turtle")
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        v = r['Location']
        self.assertEqual(v, TestHostUri+u+layout.TYPE_META_TURTLE)
        with MockHttpDictResources(v, self.get_context_mock_dict(v)):
            r = self.client.get(v)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Parse graph data as Turtle
        g = Graph()
        result = g.parse(data=r.content, publicID=v, format="turtle")
        # Check the resulting graph contents
        subj            = TestHostUri + u.rstrip("/")
        type_vocab_data = type_vocab.get_values()
        vocab_list_url  = self.resolve_coll_url(self.testcoll, type_vocab_data[ANNAL.CURIE.type_list])
        vocab_view_url  = self.resolve_coll_url(self.testcoll, type_vocab_data[ANNAL.CURIE.type_view])
        for (s, p, o) in (
            [ (subj, RDF.URI.type,        URIRef(ANNAL.URI.Type)                        )
            , (subj, RDFS.URI.label,      Literal(type_vocab_data[RDFS.CURIE.label])    )
            , (subj, RDFS.URI.comment,    Literal(type_vocab_data[RDFS.CURIE.comment])  )
            , (subj, ANNAL.URI.id,        Literal(type_vocab_data[ANNAL.CURIE.id])      )
            , (subj, ANNAL.URI.type_id,   Literal(type_vocab_data[ANNAL.CURIE.type_id]) )
            , (subj, ANNAL.URI.type_list, URIRef(vocab_list_url)                        )
            , (subj, ANNAL.URI.type_view, URIRef(vocab_view_url)                        )
            , (subj, ANNAL.URI.uri,       URIRef(ANNAL.URI.Vocabulary)                  )
            ]):
            self.assertIn( (URIRef(s), URIRef(p), o), g )
        return

    #   -----------------------------------------------------------------------------
    #   Entity list content negotiation tests
    #   -----------------------------------------------------------------------------

    def get_list_turtle(self, list_url, subj_ref, coll_id="testcoll", type_id=None, context_path=""):
        """
        Return list of entity nodes from Turtle list
        """
        self.testcoll.generate_coll_jsonld_context()
        r = self.client.get(list_url, HTTP_ACCEPT="text/turtle")
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        turtle_url        = TestHostUri + r['Location']
        expect_turtle_url = make_resource_url(TestHostUri, list_url, layout.ENTITY_LIST_TURTLE)
        self.assertEqual(turtle_url, expect_turtle_url)
        with MockHttpDictResources(turtle_url, self.get_context_mock_dict(turtle_url, context_path=context_path)):
            r = self.client.get(turtle_url)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # print("***** turtle_url: "+turtle_url)
        # print("***** c: (testcoll/_type list):")
        # print r.content
        g = Graph()
        result = g.parse(data=r.content, publicID=turtle_url, format="turtle")
        # Check the resulting graph contents
        subj = urljoin(turtle_url, subj_ref)
        for (s, p, o) in (
            [ (subj, ANNAL.URI.entity_list, None)
            ]):
            self.assertTripleIn( (URIRef(s), URIRef(p), o), g)
        list_head  = g.value(subject=URIRef(subj), predicate=URIRef(ANNAL.URI.entity_list))
        list_items = list(self.scan_rdf_list(g, list_head))
        return (turtle_url, list_items)

    def test_http_conneg_turtle_list_coll_type(self):
        """
        Content negotiate for Turtle version of a list of collection-defined types
        """
        list_url = entitydata_list_type_url(
            "testcoll", "_type",
            scope=None
            )
        subj_ref = entitydata_list_type_url(
            "testcoll", "_type",
            list_id="Type_list",
            scope=None
            )
        (turtle_url, list_items) = self.get_list_turtle(
            list_url, subj_ref, coll_id="testcoll", type_id=None, context_path="../"
            )
        t_uri = urljoin(
            turtle_url, 
            entity_uriref(coll_id="testcoll", type_id="_type", entity_id="testtype")
            )
        self.assertEqual(len(list_items), 1)
        self.assertIn(URIRef(t_uri), list_items)
        return

    def test_http_conneg_turtle_list_coll_all(self):
        """
        Content negotiate for Turtle version of a list of collection-defined entities of all types
        """
        list_url = entitydata_list_all_url("testcoll", scope=None)
        subj_ref = entitydata_list_all_url(
            "testcoll", list_id="Default_list_all", scope=None
            )
        (turtle_url, list_items) = self.get_list_turtle(
            list_url, subj_ref, coll_id="testcoll", type_id=None, context_path=""
            )
        t_uri = urljoin(
            turtle_url, 
            entity_uriref(coll_id="testcoll", type_id="_type", entity_id="testtype")
            )
        e1uri = urljoin(
            turtle_url, 
            entity_uriref(coll_id="testcoll", type_id="testtype", entity_id="entity1")
            )
        self.assertEqual(len(list_items), 2)
        self.assertIn(URIRef(t_uri), list_items)
        self.assertIn(URIRef(e1uri), list_items)
        return

    def test_http_conneg_turtle_list_coll_all_list(self):
        """
        Content negotiate for Turtle version of a list of collection-defined entities of all types
        """
        list_url = entitydata_list_all_url(
            "testcoll", list_id="Type_list", scope=None
            )
        subj_ref = entitydata_list_all_url(
            "testcoll", list_id="Type_list", scope=None
            )
        (turtle_url, list_items) = self.get_list_turtle(
            list_url, subj_ref, coll_id="testcoll", type_id=None, context_path="../../d/"
            )
        t_uri = urljoin(
            turtle_url, 
            entity_uriref(coll_id="testcoll", type_id="_type", entity_id="testtype")
            )
        e1uri = urljoin(
            turtle_url, 
            entity_uriref(coll_id="testcoll", type_id="testtype", entity_id="entity1")
            )
        self.assertEqual(len(list_items), 1)
        self.assertIn(URIRef(t_uri), list_items)
        return

    def test_http_conneg_turtle_list_site_type(self):
        """
        Content negotiate for Turtle version of a list of collection-defined types
        """
        list_url = entitydata_list_type_url(
            "testcoll", "_type",
            scope="all"
            )
        subj_ref = entitydata_list_type_url(
            "testcoll", "_type",
            list_id="Type_list",
            scope="all"
            )
        (turtle_url, list_items) = self.get_list_turtle(
            list_url, subj_ref, coll_id="testcoll", type_id=None, context_path="../"
            )
        # print "@@ "+repr(list_items)
        expect_type_ids = get_site_types()
        expect_type_ids.add("testtype")
        self.assertEqual(len(list_items), len(expect_type_ids))
        for entity_id in expect_type_ids:
            t_uri = urljoin(
                turtle_url, 
                entity_uriref(coll_id="testcoll", type_id="_type", entity_id=entity_id)
                )
            self.assertIn(URIRef(t_uri), list_items)
        return

    def test_http_conneg_turtle_list_site_all_list(self):
        """
        Content negotiate for Turtle version of a list of 
        collection-defined entities of all types
        """
        list_url = entitydata_list_all_url(
            "testcoll", list_id="Type_list", scope="all"
            )
        subj_ref = entitydata_list_all_url(
            "testcoll", list_id="Type_list", scope="all"
            )
        (turtle_url, list_items) = self.get_list_turtle(
            list_url, subj_ref, coll_id="testcoll", type_id=None, context_path="../../d/"
            )
        expect_type_ids = get_site_types()
        expect_type_ids.add("testtype")
        self.assertEqual(len(list_items), len(expect_type_ids))
        for entity_id in expect_type_ids:
            t_uri = urljoin(
                turtle_url, 
                entity_uriref(coll_id="testcoll", type_id="_type", entity_id=entity_id)
                )
            self.assertIn(URIRef(t_uri), list_items)
        return

    def test_http_conneg_turtle_list_coll_wrongname(self):
        """
        This is a test of the regular expression used to access list resources,
        to ensure the '.' must be a literal '.' and not a wildcard.
        """
        list_url = entitydata_list_type_url(
            "testcoll", "_type",
            scope=None
            )
        wrong_name     = layout.ENTITY_LIST_FILE.replace('.', '*')
        wrong_turtle_url = make_resource_url(TestHostUri, list_url, wrong_name)
        with SuppressLogging(logging.WARNING):
            r = self.client.get(wrong_turtle_url)
        self.assertEqual(r.status_code,   404)
        self.assertEqual(r.reason_phrase, "Not found")
        return

# End.



"""
Utility functions to support entity data testing
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import os

from utils.py3porting           import urljoin

from django.conf                import settings
from django.http                import QueryDict
from django.utils.http          import urlquote, urlunquote
from django.core.urlresolvers   import resolve, reverse

from annalist.util              import valid_id
from annalist.identifiers       import RDF, RDFS, ANNAL
from annalist                   import layout

from annalist.views.form_utils.fieldchoice  import FieldChoice
from annalist.views.fields.render_placement import (
    get_placement_classes
    )

from .tests import (
    TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
    )
from .entity_testutils import (
    collection_dir, 
    site_title, 
    collection_entity_view_url,
    context_field_row
    )
from .entity_testentitydata      import entitydata_list_type_url
from .entity_testfielddesc       import get_field_description, get_bound_field
from .entity_testtypedata        import recordtype_url
from .entity_testsitedata import (
    make_field_choices, no_selection,
    get_site_types, get_site_types_sorted, get_site_types_linked,
    get_site_lists, get_site_lists_sorted, get_site_lists_linked,
    get_site_views, get_site_views_sorted, get_site_views_linked,
    get_site_list_types, get_site_list_types_sorted, get_site_list_types_linked,
    get_site_field_groups, get_site_field_groups_sorted, 
    get_site_fields, get_site_fields_sorted, 
    get_site_field_types, get_site_field_types_sorted, 
    )

#   -----------------------------------------------------------------------------
#
#   Constants
#
#   -----------------------------------------------------------------------------

# Defined here to facilitate test suite changes

num_testcoll_enumerate_all_entities     = 187   # Entities defined
num_testcoll_all_entities_scope_all     = 184   # Entities listed
num_testcoll_inherit_entities_scope_all = 238   # Entities listed, including inherited


#   -----------------------------------------------------------------------------
#
#   Directory generating functions
#
#   -----------------------------------------------------------------------------

def recordlist_dir(coll_id="testcoll", list_id="testlist"):
    return collection_dir(coll_id) + layout.COLL_LIST_PATH%{'id': list_id} + "/"

#   -----------------------------------------------------------------------------
#
#   URI generating functions
#
#   -----------------------------------------------------------------------------

def recordlist_coll_url(site, coll_id="testcoll", list_id="testlist"):
    return urljoin(
        site._entityurl,
        layout.SITE_COLL_PATH%{'id': coll_id} + "/" + 
        layout.COLL_LIST_PATH%{'id': list_id} + "/"
        )

def recordlist_url(coll_id, list_id):
    """
    URI for record list description data; also view using default entity view
    """
    if not valid_id(list_id):
        list_id = "___"
    return collection_entity_view_url(coll_id=coll_id, type_id="_list", entity_id=list_id)

def recordlist_edit_url(action=None, coll_id=None, list_id=None):
    """
    URI for record list description editing view
    """
    viewname = ( 
        'AnnalistEntityDataView'        if action == "view"   else
        'AnnalistEntityNewView'         if action == "new"    else
        'AnnalistEntityEditView'        if action == "copy"   else
        'AnnalistEntityEditView'        if action == "edit"   else
        'AnnalistRecordListDeleteView'  if action == "delete" else
        'unknown'
        )
    kwargs = {'coll_id': coll_id}
    if action != "delete":
        kwargs.update({'action': action, 'type_id': "_list", 'view_id': "List_view"})
    if list_id:
        if valid_id(list_id):
            kwargs.update({'entity_id': list_id})
        else:
            kwargs.update({'entity_id': "___"})
    return reverse(viewname, kwargs=kwargs)

#   -----------------------------------------------------------------------------
#
#   ----- RecordList data
#
#   -----------------------------------------------------------------------------

def recordlist_value_keys(list_uri=False):
    keys = set(
        [ 'annal:id', 'annal:type_id'
        , 'annal:type'
        , 'annal:url'
        , 'rdfs:label', 'rdfs:comment'
        , 'annal:display_type'
        , 'annal:list_entity_selector'
        , 'annal:default_view'
        , 'annal:default_type'
        , 'annal:list_fields'
        ])
    if list_uri:
        keys.add('annal:uri')
    return keys

def recordlist_load_keys(list_uri=False):
    return recordlist_value_keys(list_uri=list_uri) | {'@id', '@type', '@context'}

def recordlist_create_values(
        coll_id="testcoll", list_id="testlist", list_uri=None, update="RecordList"):
    """
    Entity values used when creating a record list entity
    """
    d = (
        { 'annal:type':                 "annal:List"
        , 'rdfs:label':                 "%s %s/%s"%(update, coll_id, list_id)
        , 'rdfs:comment':               "%s help for %s/%s"%(update, coll_id, list_id)
        , "annal:display_type":         "_enum_list_type/List"
        , "annal:default_view":         "_view/Default_view"
        , "annal:default_type":         "_type/Default_type"
        # , "annal:list_entity_type":     None
        , "annal:list_entity_selector": "ALL"
        , "annal:list_fields":
          [ { "annal:field_id":             layout.FIELD_TYPEID+"/Entity_id"
            , "annal:field_placement":      "small:0,3"
            }
          , { "annal:field_id":             layout.FIELD_TYPEID+"/Entity_label"
            , "annal:field_placement":      "small:3,9"
            }
          ]
        })
    if list_uri:
        d['annal:uri'] = list_uri
    return d

def recordlist_values(
        coll_id="testcoll", list_id="testlist", list_uri=None,
        update="RecordList", hosturi=TestHostUri):
    list_url = recordlist_url(coll_id, list_id)
    d = recordlist_create_values(coll_id, list_id, list_uri=list_uri, update=update).copy()
    d.update(
        { 'annal:id':       list_id
        , 'annal:type_id':  "_list"
        , 'annal:url':      list_url
        })
    return d

def recordlist_read_values(
        coll_id="testcoll", list_id="testlist", 
        update="RecordList", hosturi=TestHostUri):
    d = recordlist_values(coll_id, list_id, update=update, hosturi=hosturi).copy()
    d.update(
        { '@id':            layout.COLL_BASE_LIST_REF%{'id': list_id}
        , '@type':          ["annal:List"]
        , '@context':       [{"@base": "../../"}, "../../coll_context.jsonld"]
        })
    return d

#   -----------------------------------------------------------------------------
#
#   ----- Data in recordlist view for list description data
#
#   -----------------------------------------------------------------------------

def list_view_context_data(
        coll_id="testcoll", list_id=None, orig_id=None,
        action=None,
        list_uri=None,
        list_type="List",
        list_label=None,
        list_descr=None,
        list_default_type="_type/Default_type", type_choices=None,
        list_default_view="_type/Default_view", view_choices=None,
        list_entity_selector="ALL",
        list_entity_type="",
        list_fields=None,
        num_fields=0,
        update="RecordList",
        continuation_url=None
    ):
    if list_label is None:
        if list_id:
            #@@TODO: use same format as no list_id; change form data too
            list_label = "%s %s/%s"%(update, coll_id, list_id)
        else:
            list_label = "%s list (%s/)"%(update, coll_id)
    if list_fields is None:
        if num_fields == 2:
            list_fields = (
                [ { "annal:field_id":         layout.FIELD_TYPEID+"/Entity_id"
                  , "annal:field_placement":  "small:0,3"
                  }
                , { "annal:field_id":         layout.FIELD_TYPEID+"/Entity_label"
                  , "annal:field_placement":  "small:3,9"
                  }
                ])
        if num_fields == 3:
            list_fields = (
                [ { "annal:field_id":         layout.FIELD_TYPEID+"/Entity_id"
                  , "annal:field_placement":  "small:0,3"
                  }
                , { "annal:field_id":         layout.FIELD_TYPEID+"/Entity_type" 
                  , "annal:field_placement":  "small:3,3"
                  }
                , { "annal:field_id":         layout.FIELD_TYPEID+"/Entity_label"
                  , "annal:field_placement":  "small:6,6"
                  }
                ])
    list_type_choices = get_site_list_types_linked("testcoll")
    if type_choices is None:
        type_choices = (
            [ FieldChoice("", label="(default entity type)")] +
            get_site_types_linked("testcoll") +
            [ FieldChoice("_type/testtype", 
                label="RecordType testcoll/_type/testtype",
                link=recordtype_url("testcoll", "testtype")
                )]
            )
    if view_choices is None:
        view_choices = (
            [ FieldChoice("", label="(view id)") ] +
            get_site_views_linked("testcoll")
            )
    if continuation_url is None:
        continuation_url = entitydata_list_type_url(coll_id, layout.LIST_TYPEID)
    view_label = "List definition"
    view_title = (
        "%s - %s - Collection %s"%(list_label, view_label, coll_id) if list_label
        else
        "%s - Collection %s"%(view_label, coll_id)
        )
    # Target record fields listed in the view description
    context_dict = (
        { 'title':              view_title
        , 'heading':            view_label
        , 'coll_id':            coll_id
        , 'type_id':            layout.LIST_TYPEID
        , 'view_id':            'List_view'
        , 'entity_id':          list_id or ""
        , 'orig_id':            orig_id
        , 'orig_type':          layout.LIST_TYPEID
        , 'record_type':        "annal:List"
        , 'continuation_url':   continuation_url
        , 'fields':
          [ context_field_row(
              get_bound_field("List_id",              list_id),             # 0 (0,0)
              get_bound_field("List_type",            list_type,            # 1 (0,1)
                              options=list_type_choices),
              )
          , context_field_row(
              get_bound_field("List_label",           list_label)           # 2 (1,0)
              )
          , context_field_row(
              get_bound_field("List_comment",         list_descr)           # 3 (2,0)
              )
          , context_field_row(
              get_bound_field("List_default_type",    list_default_type,    # 4 (3,0)
                              options=type_choices),
              get_bound_field("List_default_view",    list_default_view,    # 5 (3,1)
                              options=view_choices),
              )
          , context_field_row(
              get_bound_field("List_entity_selector", list_entity_selector) # 6 (4,0)
              )
          , context_field_row(
              get_bound_field("List_entity_type",     list_entity_type)     # 7 (5,0)
              )
          , get_bound_field("List_fields",            list_fields)          # 8 (6, 0)
          ]
        })
    if action:  
        context_dict['action']      = action
    if list_uri:
        context_dict['entity_uri']  = list_uri
    return context_dict

def list_view_form_data(
        coll_id="testcoll", orig_coll=None,
        list_id="", orig_id=None, 
        action=None, cancel=None, task=None,
        update="RecordView"):
    form_data_dict = (
        { "List_type":              "_enum_list_type/List"
        , "List_label":             "%s list (%s/%s)"%(update, coll_id, list_id)
        , "List_comment":           "%s help (%s/%s)"%(update, coll_id, list_id)
        , "List_default_type":      "_type/Default_type"
        , "List_default_view":      "_view/Default_view"
        , "List_entity_selector":   "ALL"
        # List repeating fields
        , "List_fields__0__Field_id":           layout.FIELD_TYPEID+"/Entity_id"
        , "List_fields__0__Field_placement":    "small:0,3"
        , "List_fields__1__Field_id":           layout.FIELD_TYPEID+"/Entity_label"
        , "List_fields__1__Field_placement":    "small:3,9"
        # Hidden fields
        , "action":                 action
        , "view_id":                "List_view"
        , "orig_id":                "orig_list_id"
        , "orig_type":              "_list"
        , "orig_coll":              coll_id
        , "continuation_url":       entitydata_list_type_url(coll_id, "_list")        
        })
    if list_id is not None:
        form_data_dict['entity_id']     = list_id
    if list_id:
        form_data_dict['entity_id']     = list_id
        form_data_dict['orig_id']       = list_id
        form_data_dict['List_label']    = "%s %s/%s"%(update, coll_id, list_id)
        form_data_dict['List_comment']  = "%s help for %s/%s"%(update, coll_id, list_id)
    if orig_id:
        form_data_dict['orig_id']       = orig_id
    if orig_coll:
        form_data_dict['orig_coll']     = orig_coll
    if action:
        form_data_dict['action']        = action
    if cancel:
        form_data_dict['cancel']        = "Cancel"
    elif task:
        form_data_dict[task]            = task        
    else:
        form_data_dict['save']          = "Save"
    return form_data_dict

#   -----------------------------------------------------------------------------
#
#   ----- Recordview delete confirmation form data
#
#   -----------------------------------------------------------------------------

def recordlist_delete_confirm_form_data(list_id=None):
    return (
        { 'listlist':    list_id,
          'list_delete': 'Delete'
        })

# End.



from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
Value mapper for Identifier (URI/CURIE) field
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2017, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

from utils.py3porting                   import is_string, to_unicode

from annalist.views.fields.render_base  import RenderBase

#   ----------------------------------------------------------------------------
#
#   Identifier value mapping
#
#   ----------------------------------------------------------------------------

class IdentifierValueMapper(RenderBase):
    """
    Value mapper class for entity id field.
    """

    @classmethod
    def decode(cls, field_value):
        """
        Returns an Identifier (URI/CURIE) form field value with leading/trailing spaces trimmed
        """
        if is_string(field_value):
            field_value = field_value.strip()
        return field_value

    # encode defaults to identity mapper]

# End.



"""
Tests for RecordVocab module and view

Note: this module tests for rendering specifically for RecordVocab values, using
type description sitedata files, and as such duplicates some tests covered by
module test_entitygenericedit.
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2017, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import os
import unittest
import markdown

from django.conf                        import settings
from django.db                          import models
from django.http                        import QueryDict
from django.contrib.auth.models         import User
from django.test                        import TestCase # cf. https://docs.djangoproject.com/en/dev/topics/testing/tools/#assertions
from django.test.client                 import Client

from annalist.util                      import valid_id, extract_entity_id
from annalist.identifiers               import RDF, RDFS, ANNAL
from annalist                           import layout
from annalist                           import message

from annalist.models.site               import Site
from annalist.models.sitedata           import SiteData
from annalist.models.collection         import Collection
from annalist.models.entitydata         import EntityData
from annalist.models.recordvocab        import RecordVocab

from annalist.views.form_utils.fieldchoice  import FieldChoice
from annalist.views.displayinfo             import apply_substitutions

from .AnnalistTestCase import AnnalistTestCase
from .tests import (
    TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
    )
from .init_tests import (
    init_annalist_test_site,
    init_annalist_test_coll,
    resetSitedata
    )
from .entity_testfielddesc import get_field_description, get_bound_field
from .entity_testutils import (
    make_message, make_quoted_message,
    site_dir, collection_dir,
    site_view_url, collection_edit_url, 
    collection_entity_view_url,
    collection_create_values,
    create_test_user, create_user_permissions,
    context_view_field,
    context_bind_fields,
    check_context_field, check_context_field_value,
    )
from .entity_testvocabdata import (
    recordvocab_dir,
    recordvocab_coll_url, recordvocab_url, recordvocab_edit_url,
    recordvocab_value_keys, recordvocab_load_keys, 
    recordvocab_create_values, recordvocab_values, recordvocab_read_values,
    vocab_view_context_data, 
    vocab_view_form_data, # recordvocab_delete_confirm_form_data
    )
from .entity_testentitydata import (
    entity_url, entitydata_edit_url, entitydata_list_type_url,
    default_fields, default_label, default_comment, error_label,
    layout_classes
    )
from .entity_testsitedata import (
    get_site_types, get_site_types_sorted, get_site_types_linked,
    get_site_lists, get_site_lists_sorted, get_site_lists_linked,
    get_site_views, get_site_views_sorted, get_site_views_linked,
    get_site_list_types, get_site_list_types_sorted,
    get_site_field_groups, get_site_field_groups_sorted, 
    get_site_fields, get_site_fields_sorted, 
    get_site_field_types, get_site_field_types_sorted, 
    )
from .entity_testviewdata import recordview_url
from .entity_testlistdata import recordlist_url

#   -----------------------------------------------------------------------------
#
#   RecordVocab tests
#
#   -----------------------------------------------------------------------------

class RecordVocabTest(AnnalistTestCase):
    """
    Tests for RecordVocab object interface
    """

    def setUp(self):
        init_annalist_test_site()
        self.testsite = Site(TestBaseUri, TestBaseDir)
        self.sitedata = SiteData(self.testsite)
        self.testcoll = Collection(self.testsite, "testcoll")
        self.layout = (
            { 'enum_field_placement_id':    layout.ENUM_FIELD_PLACEMENT_ID
            , 'enum_list_type_id':          layout.ENUM_LIST_TYPE_ID
            , 'enum_render_type_id':        layout.ENUM_RENDER_TYPE_ID
            , 'enum_value_type_id':         layout.ENUM_VALUE_TYPE_ID
            , 'enum_value_mode_id':         layout.ENUM_VALUE_MODE_ID
            , 'field_typeid':               layout.FIELD_TYPEID
            , 'group_typeid':               layout.GROUP_TYPEID
            , 'list_typeid':                layout.LIST_TYPEID
            , 'type_typeid':                layout.TYPE_TYPEID
            , 'user_typeid':                layout.USER_TYPEID
            , 'view_typeid':                layout.VIEW_TYPEID
            , 'vocab_typeid':               layout.VOCAB_TYPEID
            , 'field_dir':                  layout.FIELD_DIR
            , 'group_dir':                  layout.GROUP_DIR
            , 'list_dir':                   layout.LIST_DIR
            , 'type_dir':                   layout.TYPE_DIR
            , 'user_dir':                   layout.USER_DIR
            , 'view_dir':                   layout.VIEW_DIR
            , 'vocab_dir':                  layout.VOCAB_DIR
            })
        return

    def tearDown(self):
        return

    @classmethod
    def tearDownClass(cls):
        resetSitedata(scope="collections")
        return

    # @classmethod
    # def setUpClass(cls):
    #     super(zzzzzz, cls).setUpClass()
    #     return

    # @classmethod
    # def tearDownClass(cls):
    #     super(zzzzzz, cls).tearDownClass()
    #     return

    def test_RecordVocabTest(self):
        self.assertEqual(RecordVocab.__name__, "RecordVocab", "Check RecordVocab class name")
        return

    def test_recordvocab_init(self):
        t = RecordVocab(self.testcoll, "testvocab")
        u = recordvocab_coll_url(self.testsite, coll_id="testcoll", vocab_id="testvocab")
        self.assertEqual(t._entitytype,     ANNAL.CURIE.Vocabulary)
        self.assertEqual(t._entityfile,     layout.VOCAB_META_FILE)
        self.assertEqual(t._entityref,      layout.COLL_BASE_VOCAB_REF%{'id': "testvocab"})
        self.assertEqual(t._entityid,       "testvocab")
        self.assertEqual(t._entityurl,      u)
        self.assertEqual(t._entitydir,      recordvocab_dir(vocab_id="testvocab"))
        self.assertEqual(t._values,         None)
        return

    def test_recordvocab1_data(self):
        t = RecordVocab(self.testcoll, "vocab1")
        self.assertEqual(t.get_id(), "vocab1")
        self.assertEqual(t.get_type_id(), layout.VOCAB_TYPEID)
        self.assertIn(
            "/c/testcoll/d/%(vocab_dir)s/vocab1/"%self.layout, 
            t.get_url()
            )
        self.assertEqual(
            TestBaseUri + "/c/testcoll/d/%(vocab_typeid)s/vocab1/"%self.layout, 
            t.get_view_url()
            )
        t.set_values(recordvocab_create_values(vocab_id="vocab1"))
        td = t.get_values()
        self.assertEqual(set(td.keys()), set(recordvocab_value_keys()))
        v = recordvocab_values(vocab_id="vocab1")
        self.assertDictionaryMatch(td, v)
        return

    def test_recordvocab2_data(self):
        t = RecordVocab(self.testcoll, "vocab2")
        self.assertEqual(t.get_id(), "vocab2")
        self.assertEqual(t.get_type_id(), layout.VOCAB_TYPEID)
        self.assertIn(
            "/c/testcoll/d/%(vocab_dir)s/vocab2/"%self.layout, 
            t.get_url()
            )
        self.assertEqual(
            TestBaseUri + "/c/testcoll/d/%(vocab_typeid)s/vocab2/"%self.layout, 
            t.get_view_url()
            )
        t.set_values(recordvocab_create_values(vocab_id="vocab2"))
        td = t.get_values()
        self.assertEqual(set(td.keys()), set(recordvocab_value_keys()))
        v = recordvocab_values(vocab_id="vocab2")
        self.assertDictionaryMatch(td, v)
        return

    def test_recordvocab_create_load(self):
        t  = RecordVocab.create(self.testcoll, "vocab1", recordvocab_create_values(vocab_id="vocab1"))
        td = RecordVocab.load(self.testcoll, "vocab1").get_values()
        v  = recordvocab_read_values(vocab_id="vocab1")
        self.assertKeysMatch(td, v)
        self.assertDictionaryMatch(td, v)
        return

#   -----------------------------------------------------------------------------
#
#   RecordVocabEditView tests
#
#   -----------------------------------------------------------------------------

class RecordVocabEditViewTest(AnnalistTestCase):
    """
    Tests for vocabulary record edit views
    """

    def setUp(self):
        init_annalist_test_site()
        self.testsite   = Site(TestBaseUri, TestBaseDir)
        self.testcoll   = Collection.create(
            self.testsite, "testcoll", collection_create_values("testcoll")
            )
        self.annalcoll  = Collection.load(self.testsite, layout.SITEDATA_ID)
        self.no_options = [ FieldChoice('', label="(no options)") ]
        # For checking Location: header values...
        self.continuation_url = (
            entitydata_list_type_url(coll_id="testcoll", type_id=layout.VOCAB_TYPEID)
            )
        # Login and permissions
        create_test_user(self.testcoll, "testuser", "testpassword")
        create_user_permissions(self.annalcoll, "testuser", user_permissions=["VIEW"])
        self.client = Client(HTTP_HOST=TestHost)
        loggedin = self.client.login(username="testuser", password="testpassword")
        self.assertTrue(loggedin)
        return

    def tearDown(self):
        resetSitedata(scope="collections")
        return

    @classmethod
    def setUpClass(cls):
        super(RecordVocabEditViewTest, cls).setUpClass()
        return

    @classmethod
    def tearDownClass(cls):
        super(RecordVocabEditViewTest, cls).tearDownClass()
        # @@checkme@@ resetSitedata()
        return

    #   -----------------------------------------------------------------------------
    #   Helpers
    #   -----------------------------------------------------------------------------

    def _create_record_vocab(self, vocab_id):
        "Helper function creates namespace vocabulary entry with supplied vocab_id"
        ve = RecordVocab.create(
            self.testcoll, vocab_id, recordvocab_create_values(vocab_id=vocab_id)
            )
        return ve

    def _check_record_vocab_values(self, vocab_id, 
            update="RecordVocab", 
            vocab_uri=None,
            extra_values=None
            ):
        "Helper function checks content of record type entry with supplied vocab_id"
        self.assertTrue(RecordVocab.exists(self.testcoll, vocab_id))
        ve = RecordVocab.load(self.testcoll, vocab_id)
        self.assertEqual(ve.get_id(), vocab_id)
        self.assertEqual(ve.get_view_url(), TestHostUri + recordvocab_url("testcoll", vocab_id))
        vv = recordvocab_values(
            vocab_id=vocab_id,
            update=update,
            vocab_uri=vocab_uri
            )
        if extra_values:
            vv.update(extra_values)
        # print "ve: "+repr(ve.get_values())
        # print "vv: "+repr(v)
        self.assertDictionaryMatch(ve.get_values(), vv)
        return ve

    def _check_context_fields(self, response, 
            action="",
            vocab_id="", orig_vocab_id=None,
            vocab_label=None,
            vocab_descr=None,
            vocab_uri=None,
            vocab_seealso=[],
            update="RecordVocab",
            continuation_url=None
            ):
        expect_context = vocab_view_context_data(
            coll_id="testcoll", vocab_id=vocab_id, orig_id=orig_vocab_id, action=action, 
            vocab_label=vocab_label,
            vocab_descr=vocab_descr,
            vocab_uri=vocab_uri,
            vocab_seealso=[{"@id": v} for v in vocab_seealso],
            update=update,
            continuation_url=continuation_url
            )
        actual_context = context_bind_fields(response.context)
        self.assertEqual(len(response.context['fields']), 5)
        self.assertDictionaryMatch(actual_context, expect_context)
        return

    #   -----------------------------------------------------------------------------
    #   Form rendering tests
    #   -----------------------------------------------------------------------------

    def test_get_form_rendering(self):
        u = entitydata_edit_url("new", "testcoll", layout.VOCAB_TYPEID, view_id="Vocab_view")
        r = self.client.get(u+"?continuation_url=/xyzzy/")
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        field_vals = default_fields(
            coll_id="testcoll", type_id=layout.VOCAB_TYPEID, entity_id="00000001",
            default_label="",
            default_comment=context_view_field(r.context, 2, 0)['field_value'],
            default_label_esc="",
            default_comment_esc=context_view_field(r.context, 2, 0)['field_value'],
            tooltip1=context_view_field(r.context, 0, 0)['field_tooltip'],
            tooltip2=context_view_field(r.context, 1, 0)['field_tooltip'],
            tooltip3=context_view_field(r.context, 2, 0)['field_tooltip'],
            tooltip4=context_view_field(r.context, 3, 0)['field_tooltip'],
            button_save_tip="Save values and return to previous view.",
            button_view_tip="Save values and switch to entity view.",
            button_cancel_tip="Discard unsaved changes and return to previous view.",
            )
        formrow1 = """
            <div class="small-12 medium-6 columns" title="%(tooltip1)s">
              <div class="row view-value-row">
                <div class="%(label_classes)s">
                  <span>Prefix</span>
                </div>
                <div class="%(input_classes)s">
                  <input type="text" size="64" name="entity_id" 
                         placeholder="(vocabulary id)" value="00000001"/>
                </div>
              </div>
            </div>
            """%field_vals(width=6)
        formrow2 = """
            <div class="small-12 columns" title="%(tooltip2)s">
              <div class="row view-value-row">
                <div class="%(label_classes)s">
                  <span>Label</span>
                </div>
                <div class="%(input_classes)s">
                  <input type="text" size="64" name="Entity_label" 
                         placeholder="(label)" 
                         value="%(default_label_esc)s" />
                </div>
              </div>
            </div>
            """%field_vals(width=12)
        formrow3 = """
            <div class="small-12 columns" title="%(tooltip3)s">
              <div class="row view-value-row">
                <div class="%(label_classes)s">
                  <span>Comment</span>
                </div>
                <div class="%(input_classes)s">
                  <textarea cols="64" rows="6" name="Entity_comment" 
                            class="small-rows-4 medium-rows-8"
                            placeholder="(description)"
                            >
                      %(default_comment_esc)s
                  </textarea>
                </div>
              </div>
            </div>
            """%field_vals(width=12)
        formrow4 = """
            <div class="small-12 columns" title="%(tooltip4)s">
              <div class="row view-value-row">
                <div class="%(label_classes)s">
                  <span>Vocabulary URI</span>
                </div>
                <div class="%(input_classes)s">
                  <input type="text" size="64" name="Vocab_uri" 
                         placeholder="(Vocabulary namespace URI)"
                         value=""/>
                </div>
              </div>
            </div>
            """%field_vals(width=12)
        # log.info(r.content)   #@@
        self.assertContains(r, formrow1, html=True)
        self.assertContains(r, formrow2, html=True)
        self.assertContains(r, formrow3, html=True)
        self.assertContains(r, formrow4, html=True)
        return

    def test_get_view_rendering(self):
        # Test view rendering
        # This checks for namespace prefix expansion in URL links.
        # Add test namespace
        annal_uri    = ANNAL.mk_uri("")
        vocab_id     = "test_vocab"
        vocab_curie  = "annal:"+vocab_id+"#"
        vocab_uri    = annal_uri+vocab_id+"#"
        vocab_entity = RecordVocab.create(
            self.testcoll, vocab_id, 
            recordvocab_create_values(vocab_id=vocab_id, vocab_uri=vocab_curie)
            )
        # View new namespace entity
        u = entitydata_edit_url(
            "view", "testcoll", layout.VOCAB_TYPEID, entity_id=vocab_id, view_id="Vocab_view"
            )
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        default_comment = context_view_field(r.context, 2, 0)['field_value']
        field_vals = default_fields(
            coll_id="testcoll", type_id=layout.VOCAB_TYPEID, entity_id=vocab_id,
            entity_view_url=u,
            rendered_label=context_view_field(r.context, 1, 0)['field_value'],
            rendered_comment=markdown.markdown(apply_substitutions(r.context, default_comment)),
            vocab_id=vocab_id,
            vocab_curie=vocab_curie,
            vocab_uri=vocab_uri,
            )
        formrow1 = """
            <div class="small-12 medium-6 columns">
              <div class="row view-value-row">
                <div class="%(label_classes)s">
                  <span>Prefix</span>
                </div>
                <div class="%(input_classes)s">
                  <a href="%(default_entity_url)s?continuation_url=%(entity_view_url)s">%(vocab_id)s</a>
                </div>
              </div>
            </div>
            """%field_vals(width=6)
        formrow2 = """
            <div class="small-12 columns">
              <div class="row view-value-row">
                <div class="%(label_classes)s">
                  <span>Label</span>
                </div>
                <div class="%(input_classes)s">
                  <span>%(rendered_label)s</span>
                </div>
              </div>
            </div>
            """%field_vals(width=12)
        formrow3 = """
            <div class="small-12 columns">
              <div class="row view-value-row">
                <div class="%(label_classes)s">
                  <span>Comment</span>
                </div>
                <div class="%(input_classes)s">
                  <span class="markdown">
                    %(rendered_comment)s
                  </span>
                </div>
              </div>
            </div>
            """%field_vals(width=12)
        formrow4 = """
            <div class="small-12 columns">
              <div class="row view-value-row">
                <div class="%(label_classes)s">
                  <span>Vocabulary URI</span>
                </div>
                <div class="%(input_classes)s">
                  <a href="%(vocab_uri)s" target="_blank">%(vocab_curie)s</a>
                </div>
              </div>
            </div>
            """%field_vals(width=12)
        # log.info(r.content)   #@@
        self.assertContains(r, formrow1, html=True)
        self.assertContains(r, formrow2, html=True)
        self.assertContains(r, formrow3, html=True)
        self.assertContains(r, formrow4, html=True)
        return

    def test_get_new(self):
        u = entitydata_edit_url("new", "testcoll", layout.VOCAB_TYPEID, view_id="Vocab_view")
        r = self.client.get(u+"?continuation_url=/xyzzy/")
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Test context
        type_url = collection_entity_view_url(
            coll_id="testcoll", type_id=layout.VOCAB_TYPEID, entity_id="00000001"
            )
        self.assertEqual(r.context['coll_id'],          "testcoll")
        self.assertEqual(r.context['type_id'],          layout.VOCAB_TYPEID)
        self.assertEqual(r.context['entity_id'],        "00000001")
        self.assertEqual(r.context['orig_id'],          None)
        self.assertEqual(r.context['entity_uri'],       "")
        self.assertEqual(r.context['action'],           "new")
        self.assertEqual(r.context['continuation_url'], "/xyzzy/")
        # Fields
        self._check_context_fields(r, 
            action="new",
            vocab_id="00000001", orig_vocab_id=None,
            vocab_label="",
            vocab_descr="",
            vocab_uri="",
            continuation_url="/xyzzy/"
            )
        return

    def test_get_copy(self):
        u = entitydata_edit_url(
            "copy", "testcoll", layout.VOCAB_TYPEID, entity_id="annal", view_id="Vocab_view"
            )
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Test context (values read from test data fixture)
        type_url = collection_entity_view_url(
            coll_id="testcoll", type_id=layout.VOCAB_TYPEID, entity_id="annal"
            )
        self.assertEqual(r.context['coll_id'],          "testcoll")
        self.assertEqual(r.context['type_id'],          layout.VOCAB_TYPEID)
        self.assertEqual(r.context['entity_id'],        "annal_01")
        self.assertEqual(r.context['orig_id'],          "annal")
        self.assertEqual(r.context['entity_uri'],       None)
        self.assertEqual(r.context['action'],           "copy")
        self.assertEqual(r.context['continuation_url'], "")
        # Fields
        self._check_context_fields(r, 
            action="copy",
            vocab_id="annal_01", orig_vocab_id="annal",
            vocab_label="Vocabulary namespace for Annalist-defined terms",
            vocab_uri="",
            vocab_seealso=["https://github.com/gklyne/annalist/blob/master/src/annalist_root/annalist/identifiers.py"],
            continuation_url=""
            )
        return

    def test_get_copy_not_exists(self):
        u = entitydata_edit_url(
            "copy", "testcoll", layout.VOCAB_TYPEID, 
            entity_id="novocab", view_id="Vocab_view"
            )
        r = self.client.get(u)
        # log.info(r.content)
        self.check_entity_not_found_response(r, 
            err_msg=make_message(
                message.ENTITY_DOES_NOT_EXIST, 
                type_id=layout.VOCAB_TYPEID, 
                id="novocab", 
                label=error_label("testcoll", layout.VOCAB_TYPEID, "novocab")
                )
            )
        return

    def test_get_edit(self):
        u = entitydata_edit_url(
            "edit", "testcoll", layout.VOCAB_TYPEID, entity_id="annal", view_id="Vocab_view"
            )
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Test context (values read from test data fixture)
        type_url = collection_entity_view_url(
            coll_id="testcoll", type_id=layout.VOCAB_TYPEID, entity_id="annal"
            )
        self.assertEqual(r.context['coll_id'],          "testcoll")
        self.assertEqual(r.context['type_id'],          layout.VOCAB_TYPEID)
        self.assertEqual(r.context['entity_id'],        "annal")
        self.assertEqual(r.context['orig_id'],          "annal")
        self.assertEqual(r.context['entity_uri'],       "http://purl.org/annalist/2014/#")
        self.assertEqual(r.context['action'],           "edit")
        self.assertEqual(r.context['continuation_url'], "")
        # Fields
        self._check_context_fields(r, 
            action="edit",
            vocab_id="annal", orig_vocab_id="annal",
            vocab_label="Vocabulary namespace for Annalist-defined terms",
            vocab_uri="http://purl.org/annalist/2014/#",
            vocab_seealso=["https://github.com/gklyne/annalist/blob/master/src/annalist_root/annalist/identifiers.py"],
            continuation_url=""
            )
        return

    def test_get_edit_not_exists(self):
        u = entitydata_edit_url(
            "edit", "testcoll", layout.VOCAB_TYPEID, 
            entity_id="novocab", view_id="Vocab_view"
            )
        r = self.client.get(u)
        # log.info(r.content)
        self.check_entity_not_found_response(r, 
            err_msg=make_message(
                message.ENTITY_DOES_NOT_EXIST, 
                type_id=layout.VOCAB_TYPEID, 
                id="novocab", 
                label=error_label("testcoll", layout.VOCAB_TYPEID, "novocab")
                )
            )
        return

    #   -----------------------------------------------------------------------------
    #   Form response tests
    #   -----------------------------------------------------------------------------

    #   -------- new type --------

    def test_post_new_vocab(self):
        self.assertFalse(RecordVocab.exists(self.testcoll, "newvocab"))
        f = vocab_view_form_data(
            vocab_id="newvocab", action="new",
            vocab_uri="test:newvocab#",
            update="RecordVocab"
            )
        u = entitydata_edit_url("new", "testcoll", layout.VOCAB_TYPEID, view_id="Vocab_view")
        r = self.client.post(u, f)
        # print r.content
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        self.assertEqual(r['location'],   self.continuation_url)
        # Check that new record type exists
        self._check_record_vocab_values("newvocab", update="RecordVocab", vocab_uri="test:newvocab#")
        return

    def test_post_new_vocab_cancel(self):
        self.assertFalse(RecordVocab.exists(self.testcoll, "newvocab"))
        f = vocab_view_form_data(
            vocab_id="newvocab", action="new", cancel="Cancel", update="Updated RecordVocab"
            )
        u = entitydata_edit_url("new", "testcoll", layout.VOCAB_TYPEID, view_id="Vocab_view")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        self.assertEqual(r['location'],   self.continuation_url)
        # Check that new record type still does not exist
        self.assertFalse(RecordVocab.exists(self.testcoll, "newvocab"))
        return

    def test_post_new_vocab_missing_id(self):
        f = vocab_view_form_data(action="new", update="RecordVocab")
        u = entitydata_edit_url("new", "testcoll", layout.VOCAB_TYPEID, view_id="Vocab_view")
        r = self.client.post(u, f)
        # print r.content
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertContains(r, "<h3>%s</h3>"%(message.RECORD_VOCAB_ID))
        # Test context
        self._check_context_fields(r, 
            action="new",
            vocab_id=None, orig_vocab_id="orig_vocab_id",
            vocab_label=None,
            vocab_descr=None,
            vocab_uri=None,
            )
        return

    def test_post_new_vocab_invalid_id(self):
        f = vocab_view_form_data(
            vocab_id="!badvocab", orig_id="orig_vocab_id", action="new", update="RecordVocab"
            )
        u = entitydata_edit_url("new", "testcoll", layout.VOCAB_TYPEID, view_id="Vocab_view")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertContains(r, "<h3>%s</h3>"%(message.RECORD_VOCAB_ID))
        # Test context
        self._check_context_fields(r, 
            action="new",
            vocab_id="!badvocab", orig_vocab_id="orig_vocab_id",
            vocab_label=None,
            vocab_descr=None,
            vocab_uri=None,
            )
        return

    #   -------- copy namespace vocabulary --------

    def test_post_copy_vocab(self):
        self.assertFalse(RecordVocab.exists(self.testcoll, "copyvocab"))
        f = vocab_view_form_data(
            vocab_id="copyvocab", 
            orig_id="annal", orig_coll="_annalist_site", action="copy", 
            update="RecordVocab",
            vocab_uri=" test:copyvocab# "  # Tests space stripping
            )
        u = entitydata_edit_url(
            "copy", "testcoll", layout.VOCAB_TYPEID, entity_id="annal", view_id="Vocab_view"
            )
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        self.assertEqual(r['location'],   self.continuation_url)
        # Check that new record type exists
        self._check_record_vocab_values("copyvocab", update="RecordVocab", vocab_uri="test:copyvocab#")
        return

    def test_post_copy_vocab_cancel(self):
        self.assertFalse(RecordVocab.exists(self.testcoll, "copyvocab"))
        f = vocab_view_form_data(
            vocab_id="copyvocab", orig_id="annal", action="copy", cancel="Cancel", update="RecordVocab"
            )
        u = entitydata_edit_url(
            "copy", "testcoll", layout.VOCAB_TYPEID, entity_id="annal", view_id="Vocab_view"
            )
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        self.assertEqual(r['location'],   self.continuation_url)
        # Check that target record vocab still does not exist
        self.assertFalse(RecordVocab.exists(self.testcoll, "copytype"))
        return

    def test_post_copy_vocab_missing_id(self):
        self._create_record_vocab("Default_vocab")
        f = vocab_view_form_data(
            action="copy", 
            orig_id="Default_vocab", 
            update="Updated RecordVocab"
            )
        u = entitydata_edit_url(
            "copy", "testcoll", layout.VOCAB_TYPEID, entity_id="Default_vocab", view_id="Vocab_view"
            )
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertContains(r, "<h3>%s</h3>"%(message.RECORD_VOCAB_ID))
        # Test context
        self._check_context_fields(r, 
            action="copy",
            vocab_id="", orig_vocab_id="Default_vocab",
            update="Updated RecordVocab"
            )
        return

    def test_post_copy_vocab_invalid_id(self):
        self._create_record_vocab("Default_vocab")
        f = vocab_view_form_data(
            action="copy", 
            vocab_id="!badvocab", orig_id="Default_vocab", 
            update="Updated RecordVocab"
            )
        u = entitydata_edit_url(
            "copy", "testcoll", layout.VOCAB_TYPEID, entity_id="Default_vocab", view_id="Vocab_view"
            )
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertContains(r, "<h3>%s</h3>"%(message.RECORD_VOCAB_ID))
        # Test context
        self._check_context_fields(r, 
            action="copy",
            vocab_id="!badvocab", orig_vocab_id="Default_vocab",
            update="Updated RecordVocab"
            )
        return

    #   -------- edit type --------

    def test_post_edit_vocab(self):
        self._create_record_vocab("editvocab")
        self._check_record_vocab_values("editvocab")
        f = vocab_view_form_data(
            vocab_id="editvocab", orig_id="editvocab", 
            action="edit", update="Updated RecordVocab"
            )
        u = entitydata_edit_url("edit", "testcoll", layout.VOCAB_TYPEID, entity_id="editvocab", view_id="Vocab_view")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        self.assertEqual(r['location'],   self.continuation_url)
        # Check that new record type exists
        self._check_record_vocab_values("editvocab", update="Updated RecordVocab")
        return

    def test_post_edit_vocab_cancel(self):
        self._create_record_vocab("editvocab")
        self._check_record_vocab_values("editvocab")
        f = vocab_view_form_data(
            vocab_id="editvocab", orig_id="editvocab", 
            action="edit", cancel="Cancel", update="Updated RecordVocab"
            )
        u = entitydata_edit_url(
            "edit", "testcoll", layout.VOCAB_TYPEID, entity_id="editvocab", view_id="Vocab_view"
            )
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        self.assertEqual(r['location'],   self.continuation_url)
        # Check that target record type still does not exist and unchanged
        self._check_record_vocab_values("editvocab")
        return

    def test_post_edit_vocab_missing_id(self):
        self._create_record_vocab("editvocab")
        self._check_record_vocab_values("editvocab")
        # Form post with ID missing
        f = vocab_view_form_data(
            action="edit",
            orig_id="editvocab", 
            update="Updated RecordVocab"
            )
        u = entitydata_edit_url(
            "edit", "testcoll", layout.VOCAB_TYPEID, entity_id="editvocab", view_id="Vocab_view"
            )
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertContains(r, "<h3>%s</h3>"%(message.RECORD_VOCAB_ID))
        # Test context for re-rendered form
        self._check_context_fields(r, 
            action="edit",
            vocab_id="", orig_vocab_id="editvocab",
            update="Updated RecordVocab"
            )
        # Check original data is unchanged
        self._check_record_vocab_values("editvocab")
        return

    def test_post_edit_vocab_invalid_id(self):
        self._create_record_vocab("editvocab")
        self._check_record_vocab_values("editvocab")
        # Form post with invalid ID
        f = vocab_view_form_data(
            vocab_id="!badvocab", orig_id="editvocab", action="edit", update="Updated RecordVocab"
            )
        u = entitydata_edit_url(
            "edit", "testcoll", layout.VOCAB_TYPEID, entity_id="editvocab", view_id="Vocab_view"
            )
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertContains(r, "<h3>%s</h3>"%(message.RECORD_VOCAB_ID))
        # Test context
        self._check_context_fields(r, 
            action="edit",
            vocab_id="!badvocab", orig_vocab_id="editvocab",
            update="Updated RecordVocab"
            )
        # Check original data is unchanged
        self._check_record_vocab_values("editvocab")
        return

# End.



"""
Annalist data record for a member of an enumerated type
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import os
import os.path
import shutil

from django.conf import settings

from utils.py3porting           import isoformat_space, text_to_str

from annalist                   import layout
from annalist.exceptions        import Annalist_Error
from annalist.identifiers       import ANNAL
from annalist                   import util

from annalist.models.entity     import Entity
from annalist.models.entitydata import EntityData

class RecordEnumBase(EntityData):

    _entitytypeid   = "_enum_base_id"
    _entitytype     = ANNAL.CURIE.Enum
    _entityroot     = layout.COLL_ENUM_PATH
    _entityview     = layout.COLL_ENUM_VIEW
    _entityfile     = layout.ENUM_META_FILE
    _entityprov     = layout.ENUM_PROV_FILE

    def __init__(self, parent, entity_id, type_id):
        # print("@@ RecordEnumBase.__init__ parentid %s, entityid %s"%(parent.get_id(), entity_id))
        self._entitytypeid = type_id
        super(RecordEnumBase, self).__init__(parent, entity_id)
        return

    def _migrate_filenames(self):
        """
        Override EntityData method
        """
        return None

def RecordEnumFactory(name, type_id):
    """
    Returns a dynamically-subclassed instance of RecordEnumBase using the supplied 
    class name and type_id for all created instances.
    """
    # print("@@ RecordEnumFactory name %s, type_id %s"%(name, type_id))
    def RecordEnumInit(self, parent, entity_id):
        # print("@@ RecordEnumInit parentid %s, entityid %s"%(parent.get_id(), entity_id))
        super(RecordEnumBase, self).__init__(parent, entity_id)
        return
    return type(text_to_str(name), (RecordEnumBase,), 
        { '_entitytypeid':  type_id
        , '_entityroot':    layout.COLL_ENUM_PATH%{'id': "%(id)s", 'type_id': type_id}
        , '__init__':       RecordEnumInit}
        )

# End.
